import os
import copy
import typing
import asyncio
import traceback
import importlib
from types import ModuleType

from a2wsgi import WSGIMiddleware
from starlette.types import Scope, Receive, Send, ASGIApp, Message
from starlette.status import WS_1001_GOING_AWAY
from starlette.requests import URL
from starlette.websockets import WebSocketClose
from starlette.staticfiles import StaticFiles
from starlette.responses import RedirectResponse
from starlette.middleware import Middleware
from starlette.middleware.errors import ServerErrorMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware
from starlette.exceptions import HTTPException, ExceptionMiddleware
from starlette.routing import NoMatchFound

from .types import WSGIApp
from .utils import Singleton
from .config import here, Config
from .http.request import Request
from .http.responses import FileResponse, TemplateResponse, Jinja2Templates
from .http.background import (
    BackgroundTasks,
    after_response_tasks_var,
    finished_response_tasks_var,
)
from .websocket.request import WebSocket
from .preset import (
    check_on_startup,
    clear_check_on_shutdown,
    create_directories,
    clear_directories,
)


class Lifespan:
    def __init__(
        self,
        on_startup: typing.Dict[str, typing.Callable] = None,
        on_shutdown: typing.Dict[str, typing.Callable] = None,
    ) -> None:
        self.on_startup: typing.Dict[str, typing.Callable] = on_startup or {}
        self.on_shutdown: typing.Dict[str, typing.Callable] = on_shutdown or {}

    def add_event_handler(self, event_type: str, func: typing.Callable) -> None:
        if event_type == "startup":
            self.on_startup[func.__qualname__] = func
        elif event_type == "shutdown":
            self.on_shutdown[func.__qualname__] = func
        else:
            raise ValueError("event_type must be in ('startup', 'shutdown')")

    async def startup(self) -> None:
        """
        Run any `.on_startup` event handlers.
        """
        for handler in self.on_startup.values():
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    async def shutdown(self) -> None:
        """
        Run any `.on_shutdown` event handlers.
        """
        for handler in self.on_shutdown.values():
            if asyncio.iscoroutinefunction(handler):
                await handler()
            else:
                handler()

    async def lifespan(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        message = await receive()
        assert message["type"] == "lifespan.startup"

        try:
            await self.startup()
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": msg})
            raise

        await send({"type": "lifespan.startup.complete"})
        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        await self.shutdown()
        await send({"type": "lifespan.shutdown.complete"})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "lifespan"
        await self.lifespan(scope, receive, send)


class IndexFile:
    HTTPClass = Request
    WebSocketClass = WebSocket

    def __init__(self, module_name: str, basepath: str, try_html: bool = False) -> None:
        self.module_name = module_name
        self.basepath = basepath
        self.try_html = try_html

    def _split_path(self, path: str) -> typing.List[str]:
        """
        convert url path to file string corresponding to index.py
        """
        if path.endswith("/"):
            path += "index"

        filepath = path.strip("./")
        filepath = filepath.replace("-", "_")

        pathlist = filepath.split("/")
        pathlist.insert(0, self.module_name)
        return pathlist

    def get_filepath_from_path(self, path: str) -> str:
        pathlist = self._split_path(path)
        abspath = os.path.join(self.basepath, *pathlist) + ".py"
        return abspath

    def get_module_name_from_path(self, path: str) -> typing.Optional[str]:
        """
        translate url path to module name

        if file not found, return None
        """
        pathlist = self._split_path(path)
        abspath = self.get_filepath_from_path(path)
        if not os.path.exists(abspath):
            return None
        return ".".join(pathlist)

    def get_path_from_module_name(self, module_name: str) -> typing.Optional[str]:
        """
        translate module name to url path

        if module not in base module, return None
        """
        if not module_name.startswith(self.module_name):
            return None

        path = "/".join(module_name[len(self.module_name) :].split("."))
        if path.endswith("/index"):
            path = path[:-5]
        return path

    def get_path_from_filepath(self, filepath: str) -> typing.Optional[str]:
        """
        translate file abspath to url path
        """
        assert filepath.endswith(".py")

        relpath = os.path.relpath(
            filepath, os.path.join(self.basepath, self.module_name)
        )
        if relpath.startswith("."):
            return None

        path = "/" + relpath.replace("\\", "/")[:-3]

        if path.endswith("/index"):
            path = path[:-5]

        return path

    def get_view(self, path: str) -> typing.Optional[ModuleType]:
        """
        get module from url path
        """
        module_name = self.get_module_name_from_path(path)
        if module_name is None:
            return None

        return importlib.import_module(module_name)

    def get_views(self) -> typing.Iterator[typing.Tuple[ModuleType, str]]:
        """
        return all (Module, uri)
        """
        views_path = os.path.join(self.basepath, self.module_name)

        for root, _, files in os.walk(views_path):
            try:
                files.remove("index.py")
                files.insert(0, "index.py")
            except ValueError:  # file not exists
                pass

            for file in filter(
                lambda file: file.endswith(".py") and file != "__init__.py", files
            ):
                abspath = os.path.join(root, file)
                path = self.get_path_from_filepath(abspath)
                if path is None:
                    continue
                module = self.get_view(path)
                if module is None:
                    continue
                yield module, path

    async def http(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.HTTPClass(scope, receive, send)
        pathlist = self._split_path(request.url.path)

        module = self.get_view(request.url.path)
        if not hasattr(module, "HTTP"):
            if self.try_html:
                pathlist = pathlist[1:]  # delete self.module_name from pathlist
                try:
                    # only html, no middleware/background tasks or other anything
                    await TemplateResponse(
                        os.path.join(*pathlist) + ".html", {"request": request}
                    )(scope, receive, send)
                    return
                except LookupError:
                    pass
            raise NoMatchFound()

        get_response = getattr(module, "HTTP")

        try:
            # set background tasks contextvar
            after_response_tasks_token = after_response_tasks_var.set(BackgroundTasks())
            finished_response_tasks_token = finished_response_tasks_var.set(
                BackgroundTasks()
            )
            # call middleware
            for deep in range(len(pathlist), 0, -1):
                module = importlib.import_module(".".join(pathlist[:deep]))
                if not hasattr(module, "Middleware"):
                    continue
                get_response = getattr(module, "Middleware")(get_response)

            # get response
            response = await get_response(request)
            response.background = after_response_tasks_var.get()
            await response(scope, receive, send)
        finally:
            after_response_tasks_var.reset(after_response_tasks_token)

            run_finished_response_tasks = finished_response_tasks_var.get()
            finished_response_tasks_var.reset(finished_response_tasks_token)
            await run_finished_response_tasks()

    async def websocket(self, scope: Scope, receive: Receive, send: Send) -> None:
        websocket = self.WebSocketClass(scope, receive=receive, send=send)

        module = self.get_view(websocket.url.path)
        if not hasattr(module, "Socket"):
            raise NoMatchFound()

        await getattr(module, "Socket")(websocket)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        handler = getattr(self, scope["type"])

        if scope["type"] in ("http", "websocket"):
            url = URL(scope=scope)
            path = url.path
            if path.endswith("/index"):
                handler = RedirectResponse(
                    url.replace(path=f'{path[:-len("/index")]}'), status_code=301,
                )
            elif "_" in path and not Config().ALLOW_UNDERLINE:
                handler = RedirectResponse(
                    url.replace(path=f'{path.replace("_", "-")}'), status_code=301,
                )
            elif path == "/favicon.ico":
                if os.path.exists(os.path.normpath("favicon.ico")):
                    handler = FileResponse("favicon.ico")

        await handler(scope, receive, send)


class Index(metaclass=Singleton):
    def __init__(self) -> None:
        config = Config()

        self.indexfile = IndexFile("views", here, True)
        self.templates = Jinja2Templates(config.TEMPLATES)
        self.lifespan = Lifespan(
            on_startup={
                check_on_startup.__qualname__: check_on_startup,
                create_directories.__qualname__: create_directories,
            },
            on_shutdown={
                clear_check_on_shutdown.__qualname__: clear_check_on_shutdown,
                clear_directories.__qualname__: clear_directories,
            },
        )
        self.mount_apps: typing.List[typing.Tuple[str, ASGIApp]] = [
            (
                "/static",
                StaticFiles(directory=os.path.join(here, "static"), check_dir=False),
            ),
        ]
        self.user_middlewares: typing.List[Middleware] = [
            Middleware(
                CORSMiddleware,
                allow_origins=config.CORS_ALLOW_ORIGINS,
                allow_methods=config.CORS_ALLOW_METHODS,
                allow_headers=config.CORS_ALLOW_HEADERS,
                allow_credentials=config.CORS_ALLOW_CREDENTIALS,
                allow_origin_regex=config.CORS_ALLOW_ORIGIN_REGEX,
                expose_headers=config.CORS_EXPOSE_HEADERS,
                max_age=config.CORS_MAX_AGE,
            ),
        ]
        self.exception_handlers: typing.Dict[
            typing.Union[int, typing.Type[Exception]], typing.Callable
        ] = {}

        self.asgiapp: ASGIApp = self.build_app()

    def rebuild_app(self) -> None:
        self.asgiapp = self.build_app()

    def build_app(self) -> ASGIApp:
        config = Config()
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        middlewares = [
            Middleware(TrustedHostMiddleware, allowed_hosts=config.ALLOWED_HOSTS),
            Middleware(
                ServerErrorMiddleware, handler=error_handler, debug=config.DEBUG
            ),
        ] + self.user_middlewares

        if config.FORCE_SSL:
            middlewares.append(Middleware(HTTPSRedirectMiddleware))

        middlewares += [
            Middleware(
                ExceptionMiddleware, handlers=exception_handlers, debug=config.DEBUG
            ),
        ]

        app = self.app

        for cls, options in reversed(middlewares):
            app = cls(app=app, **options)
        return app

    def add_middleware(self, middleware_class: type, **options: typing.Any) -> None:
        self.user_middlewares.insert(0, Middleware(middleware_class, **options))
        self.rebuild_app()

    def add_exception_handler(
        self,
        exc_class_or_status_code: typing.Union[int, typing.Type[Exception]],
        handler: typing.Callable,
    ) -> None:
        self.exception_handlers[exc_class_or_status_code] = handler
        self.rebuild_app()

    def exception_handler(
        self, exc_class_or_status_code: typing.Union[int, typing.Type[Exception]]
    ) -> typing.Callable:
        def decorator(func: typing.Callable) -> typing.Callable:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def on_startup(self, func: typing.Callable) -> typing.Callable:
        self.lifespan.add_event_handler("startup", func)
        return func

    def on_shutdown(self, func: typing.Callable) -> typing.Callable:
        self.lifespan.add_event_handler("shutdown", func)
        return func

    def mount_asgi(self, route: str, app: ASGIApp) -> None:
        """
        mount ASGI app
        """
        if route != "":  # allow use "" to mount app
            assert route.startswith("/"), "route must be start with '/'"
            assert not route.endswith("/"), "route can't end with '/'"
        self.mount_apps.append((route, app))

    def mount_wsgi(self, route: str, app: WSGIApp) -> None:
        """
        mount WSGI app
        """
        if route != "":  # allow use "" to mount app
            assert route.startswith("/"), "route must be start with '/'"
            assert not route.endswith("/"), "route can't end with '/'"
        self.mount_apps.append((route, WSGIMiddleware(app)))

    def mount(
        self, route: str, app: typing.Union[ASGIApp, WSGIApp], app_type: str
    ) -> None:
        import warnings

        warnings.warn(
            "`mount` method has been deprecated and will be removed in version 0.11.0 !",
            DeprecationWarning,
            stacklevel=2,
        )
        assert app_type in ("asgi", "wsgi")
        if route != "":  # allow use "" to mount app
            assert route.startswith("/"), "prefix must be start with '/'"
            assert not route.endswith("/"), "prefix can't end with '/'"
        if app_type == "wsgi":
            app = WSGIMiddleware(app)
        app = typing.cast(ASGIApp, app)
        self.mount_apps.append((route, app))

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] in ("http", "websocket"):
            path = scope["path"]
            root_path = scope.get("root_path", "")
            has_received = False

            async def subreceive() -> Message:
                nonlocal has_received
                has_received = True
                return await receive()

            async def subsend(message: Message) -> None:
                if (
                    message["type"] == "http.response.start"
                    and message["status"] == 404
                ):
                    raise NoMatchFound()
                await send(message)

            # Call into a submounted app, if one exists.
            for path_prefix, app in self.mount_apps:
                if not path.startswith(path_prefix + "/"):
                    continue
                if isinstance(app, WSGIMiddleware):
                    if scope["type"] != "http":
                        continue
                subscope = copy.copy(scope)
                subscope["path"] = path[len(path_prefix) :]
                subscope["root_path"] = root_path + path_prefix
                try:
                    await app(subscope, subreceive, subsend)
                    return
                except NoMatchFound:
                    if has_received:  # has call received
                        raise HTTPException(404)
                except HTTPException as e:
                    if e.status_code != 404 or has_received:
                        raise e

            try:
                await self.indexfile(scope, receive, send)
            except NoMatchFound:
                if scope["type"] == "http":
                    raise HTTPException(404)
                await WebSocketClose(WS_1001_GOING_AWAY)(scope, receive, send)

        elif scope["type"] == "lifespan":
            await self.lifespan(scope, receive, send)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self
        await self.asgiapp(scope, receive, send)



import re
import os
import time
import logging
import typing
import threading
import importlib
from types import ModuleType

from watchdog.events import FileSystemEventHandler, FileSystemEvent
from watchdog.observers import Observer

from .config import Config

logging.getLogger("watchdog").setLevel(logging.INFO)
logger = logging.getLogger(__name__)


IMPORT_PATTERN = re.compile("from (?P<name>.*?) import ")


class ReloadImport:
    def __init__(self) -> None:
        self.import_mapping: typing.Dict[str, typing.Set[str]] = {}
        here = os.getcwd()
        for name in os.listdir(here):
            abspath = os.path.join(here, name)
            if os.path.isdir(abspath):
                self._parse_dir(abspath)
            elif os.path.isfile(abspath):
                self._parse_file(abspath)

    def _parse_dir(self, dirpath: str) -> None:
        all_name = os.listdir(dirpath)
        if "__init__.py" not in all_name:
            return
        for name in all_name:
            abspath = os.path.join(dirpath, name)
            if os.path.isdir(abspath):
                self._parse_dir(abspath)
            if not os.path.isfile(abspath):
                continue
            self._parse_file(abspath)

    def _parse_file(self, abspath: str) -> None:
        if not abspath.endswith(".py"):
            return None

        _module_ = self.module_name(abspath)
        with open(abspath, encoding="UTF-8") as file:
            for module_name in IMPORT_PATTERN.findall(file.read()):
                if module_name in self.import_mapping:
                    self.import_mapping[module_name].add(_module_)
                else:
                    self.import_mapping[module_name] = {
                        _module_,
                    }

    @staticmethod
    def module_name(abspath: str) -> str:
        """
        Get module name from a abspath.
        """
        assert abspath.endswith(".py"), abspath
        relpath = os.path.relpath(abspath, os.getcwd()).replace("\\", "/")[:-3]
        if relpath.endswith("/__init__"):
            relpath = relpath[: -len("/__init__")]
        return relpath.replace("/", ".")

    def _import(
        self, abspath: str, *, sleep: bool = True
    ) -> typing.Optional[ModuleType]:
        """
        import module

        return: module
        """
        if sleep:  # when VSCode formats code, a temporary file will be created.
            time.sleep(1.3)
            if not os.path.exists(abspath):
                return None

        self._parse_file(abspath)

        try:
            return importlib.import_module(self.module_name(abspath))
        except SyntaxError:
            logger.debug(f"load fail {abspath}")
        return None

    def _reload(self, abspath: str) -> typing.Optional[ModuleType]:
        """
        reload module

        return: module
        """

        _module_ = self._import(abspath, sleep=False)
        if _module_:
            importlib.reload(_module_)

        for _module_name in self.import_mapping.get(self.module_name(abspath), set()):
            _abspath = os.path.join(os.getcwd(), *_module_name.split("."))
            if os.path.isdir(_abspath):
                _abspath = os.path.join(_abspath, "__init__")
            _abspath += ".py"
            self._reload(_abspath)

        return _module_


class MonitorFileEventHandler(FileSystemEventHandler):
    def __init__(self) -> None:
        self.reload_import = ReloadImport()

    def dispatch(self, event: FileSystemEvent):
        config = Config()
        if not config.HOTRELOAD or config.AUTORELOAD:
            return
        if not event.src_path.endswith(".py"):
            return
        return super().dispatch(event)

    def on_modified(self, event: FileSystemEvent):
        logger.debug(f"reloading {event.src_path}")
        threading.Thread(
            target=self.reload_import._reload, args=(event.src_path,), daemon=True
        ).start()

    def on_created(self, event: FileSystemEvent):
        logger.debug(f"loading {event.src_path}")
        threading.Thread(
            target=self.reload_import._import, args=(event.src_path,), daemon=True
        ).start()


class MonitorFile:
    def __init__(self, path: str):
        self.observer = Observer()
        self.observer.schedule(MonitorFileEventHandler(), path, recursive=True)
        self.observer.daemon = True
        self.observer.start()

    def stop(self):
        """drop observer"""
        self.observer.stop()
        self.observer.join()



import os
import sys
from copy import deepcopy
from inspect import signature
from typing import List, Dict, Any, Sequence

if sys.version_info[:2] < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict

from starlette.types import Scope, Receive, Send
from starlette.endpoints import Request, Response
from starlette.routing import NoMatchFound

from ..http.responses import (
    JSONResponse,
    YAMLResponse,
    HTMLResponse,
)
from ..applications import Index, IndexFile

from .schema import schema_parameters, schema_request_body, schema_response


class Tag(TypedDict):
    description: str
    paths: Sequence[str]


class OpenAPI:
    def __init__(
        self,
        title: str,
        description: str,
        version: str,
        *,
        tags: Dict[str, Tag] = {},
        template: str = "",
        media_type="yaml",
    ):
        """
        media_type: yaml or json
        """
        assert media_type in ("yaml", "json"), "media_type must in 'yaml' or 'json'"

        self.html_template = template
        self.media_type = media_type

        info = {"title": title, "description": description, "version": version}
        self.openapi = {
            "openapi": "3.0.0",
            "info": info,
            "paths": {},
            "tags": [
                {"name": tag_name, "description": tag_info.get("description", "")}
                for tag_name, tag_info in tags.items()
            ],
        }
        self.path2tag: Dict[str, List[str]] = {}
        for tag_name, tag_info in tags.items():
            for path in tag_info["paths"]:
                if path in self.path2tag:
                    self.path2tag[path].append(tag_name)
                else:
                    self.path2tag[path] = [tag_name]

    def _generate_paths(self) -> Dict[str, Any]:
        result = {}
        for app in reversed(
            [Index().indexfile]
            + [app for _, app in Index().mount_apps if isinstance(app, IndexFile)]
        ):
            for view, path in app.get_views():
                if not hasattr(view, "HTTP"):
                    continue
                viewclass = getattr(view, "HTTP")
                path_docs = self._generate_path(viewclass, path)
                if path_docs:
                    result[path] = path_docs
        return result

    def _generate_path(self, viewclass: object, path: str) -> Dict[str, Any]:
        result = {}
        for method in viewclass.allowed_methods():  # type: ignore
            if method == "OPTIONS":
                continue
            method = method.lower()
            method_docs = self._generate_method(viewclass, path, method)
            if method_docs:
                result[method] = method_docs
        return result

    def _generate_method(
        self, viewclass: object, path: str, method: str
    ) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        doc = getattr(viewclass, method).__doc__
        if isinstance(doc, str):
            doc = doc.strip()
            result.update(
                {
                    "summary": doc.splitlines()[0],
                    "description": "\n".join(doc.splitlines()[1:]).strip(),
                }
            )

        params = getattr(getattr(viewclass, method), "__params__", {})
        result["parameters"] = schema_parameters(
            None, params.get("query"), params.get("header"), params.get("cookie"),
        )
        if not result["parameters"]:
            del result["parameters"]

        result["requestBody"] = schema_request_body(params.get("body"))
        if not result["requestBody"]:
            del result["requestBody"]

        try:
            resps = getattr(getattr(viewclass, method), "__resps__")
        except AttributeError:
            pass
        else:
            result["responses"] = {}
            for status, content in resps.items():
                result["responses"][status] = {
                    "description": content["description"],
                }
                if content["model"] is not None:
                    result["responses"][status]["content"] = schema_response(
                        content["model"]
                    )
            if not result["responses"]:
                del result["responses"]

        if result and path in self.path2tag:  # has openapi docs, add tags
            result["tags"] = self.path2tag[path]

        return result

    def create_docs(self, request: Request) -> dict:
        openapi: dict = deepcopy(self.openapi)
        openapi["servers"] = [
            {
                "url": f"{request.url.scheme}://{request.url.netloc}",
                "description": "Current server",
            }
        ]
        openapi["paths"] = self._generate_paths()

        return openapi

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http" or scope["path"] not in ("/", "/get"):
            raise NoMatchFound()
        request = Request(scope, receive, send)

        if scope["path"] == "/get":
            handler = getattr(self, "docs")
        elif scope["path"] == "/":
            handler = getattr(self, "template")
        else:
            raise NoMatchFound()
        response = await handler(request)
        await response(scope, receive, send)

    async def template(self, request: Request) -> Response:
        if self.html_template:
            return HTMLResponse(self.html_template)

        with open(
            os.path.join(os.path.dirname(os.path.abspath(__file__)), "template.html")
        ) as file:
            DEFAULT_TEMPLATE = file.read()
        return HTMLResponse(DEFAULT_TEMPLATE)

    async def docs(self, request: Request) -> Response:
        openapi = self.create_docs(request)
        media_type = request.query_params.get("type") or self.media_type

        if media_type == "json":
            return JSONResponse(openapi)
        return YAMLResponse(openapi)



import os
import ast
import sys
import typing

import click
from uvicorn.importer import import_from_string
from starlette.testclient import TestClient as _TestClient, ASGI2App, ASGI3App
from requests import Response, Session

from .applications import Index
from .config import here, LOG_LEVELS, Config


class TestClient:
    def __init__(
        self,
        app: typing.Union[ASGI2App, ASGI3App],
        uri: str = "",
        base_url: str = "http://testserver",
        raise_server_exceptions: bool = True,
        root_path: str = "",
    ) -> None:
        self.uri = uri
        self.__client = _TestClient(app, base_url, raise_server_exceptions, root_path)

    def get(self, *args, **kwargs) -> Response:
        return self.__client.get(self.uri, *args, **kwargs)

    def options(self, *args, **kwargs) -> Response:
        return self.__client.options(self.uri, *args, **kwargs)

    def head(self, *args, **kwargs) -> Response:
        return self.__client.head(self.uri, *args, **kwargs)

    def post(self, *args, **kwargs) -> Response:
        return self.__client.post(self.uri, *args, **kwargs)

    def put(self, *args, **kwargs) -> Response:
        return self.__client.put(self.uri, *args, **kwargs)

    def patch(self, *args, **kwargs) -> Response:
        return self.__client.patch(self.uri, *args, **kwargs)

    def delete(self, *args, **kwargs) -> Response:
        return self.__client.delete(self.uri, *args, **kwargs)

    def websocket_connect(
        self, subprotocols: typing.Sequence[str] = None, **kwargs: typing.Any
    ) -> typing.Any:
        return self.__client.websocket_connect(
            self.uri, subprotocols=subprotocols, **kwargs
        )

    def __enter__(self) -> Session:
        return self.__client.__enter__()

    def __exit__(self, *args: typing.Any) -> None:
        return self.__client.__exit__()


class TestView:
    @property
    def client(self) -> TestClient:
        app = Index()
        path = app.indexfile.get_path_from_module_name(self.__class__.__module__)
        if path is None:
            raise Exception("What's wrong with you?")
        return TestClient(app, path)


class LiteralOption(click.Option):
    def type_cast_value(self, ctx, value):
        try:
            return ast.literal_eval(value)
        except:
            raise click.BadParameter(value)


@click.option(
    "-app",
    "--application",
    default=lambda: Config().APP,
    help="ASGI Application, like: main:app",
)
@click.option("--args", cls=LiteralOption, default="[]")
@click.argument("path", default="")
def cmd_test(application: str, args: typing.List[str], path: str):
    import pytest

    app: Index = import_from_string(application)
    pytest_args = [
        f"--rootdir={os.getcwd()}",
        "--override-ini=python_files=views/*.py",
        "--override-ini=python_classes=Test",
        "--override-ini=python_functions=test_*",
    ]
    pytest_args.extend(args)
    if path:
        if ".py" in path:
            pytest_args.append(path)
        elif "::" in path:
            pathlist = path.split("::")
            pathlist[0] = app.indexfile.get_filepath_from_path(pathlist[0])
            pytest_args.append("::".join(pathlist))
        else:
            pytest_args.append(app.indexfile.get_filepath_from_path(path))

    with _TestClient(app):
        sys.exit(pytest.main(pytest_args))



import asyncio
import json
import http
from typing import (
    cast,
    Any,
    Dict,
    Union,
    List,
    Iterator,
    Mapping,
    AsyncGenerator,
)

from starlette.types import Scope, Send, Receive
from starlette.datastructures import (
    Address,
    URL,
    Headers,
    QueryParams,
    State,
)
from starlette.formparsers import (
    FormData,
    MultiPartParser,
    FormParser,
    parse_options_header,
)
from starlette.requests import (
    SERVER_PUSH_HEADERS_TO_COPY,
    empty_receive,
    empty_send,
    ClientDisconnect,
)


class HTTPConnection(Mapping):
    """
    A base class for incoming HTTP connections, that is used to provide
    any functionality that is common to both `Request` and `WebSocket`.
    """

    def __init__(self, scope: Scope) -> None:
        assert scope["type"] in ("http", "websocket")
        self.scope = scope

    def __getitem__(self, key: str) -> str:
        return self.scope[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.scope)

    def __len__(self) -> int:
        return len(self.scope)

    @property
    def url(self) -> URL:
        if not hasattr(self, "_url"):
            self._url = URL(scope=self.scope)
        return self._url

    @property
    def base_url(self) -> URL:
        if not hasattr(self, "_base_url"):
            base_url_scope = dict(self.scope)
            base_url_scope["path"] = "/"
            base_url_scope["query_string"] = b""
            base_url_scope["root_path"] = base_url_scope.get(
                "app_root_path", base_url_scope.get("root_path", "")
            )
            self._base_url = URL(scope=base_url_scope)
        return self._base_url

    @property
    def headers(self) -> Headers:
        if not hasattr(self, "_headers"):
            self._headers = Headers(scope=self.scope)
        return self._headers

    @property
    def query_params(self) -> QueryParams:
        if not hasattr(self, "_query_params"):
            self._query_params = QueryParams(self.scope["query_string"])
        return self._query_params

    @property
    def cookies(self) -> Dict[str, str]:
        if not hasattr(self, "_cookies"):
            cookies = {}
            cookie_header = self.headers.get("cookie")
            if cookie_header:
                cookie = http.cookies.SimpleCookie()  # type: http.cookies.BaseCookie
                cookie.load(cookie_header)
                for key, morsel in cookie.items():
                    cookies[key] = morsel.value
            self._cookies = cookies
        return self._cookies

    @property
    def client(self) -> Address:
        host, port = self.scope.get("client") or (None, None)
        return Address(host=host, port=port)

    @property
    def state(self) -> State:
        if not hasattr(self, "_state"):
            # Ensure 'state' has an empty dict if it's not already populated.
            self.scope.setdefault("state", {})
            # Create a state instance with a reference to the dict in which it should store info
            self._state = State(self.scope["state"])
        return self._state


class Request(HTTPConnection):
    def __init__(
        self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send
    ):
        super().__init__(scope)
        assert scope["type"] == "http"
        self._receive = receive
        self._send = send
        self._stream_consumed = False
        self._is_disconnected = False

    @property
    def method(self) -> str:
        return self.scope["method"]

    @property
    def receive(self) -> Receive:
        return self._receive

    async def stream(self) -> AsyncGenerator[bytes, None]:
        if hasattr(self, "_body"):
            yield self._body
            yield b""
            return

        if self._stream_consumed:
            raise RuntimeError("Stream consumed")

        self._stream_consumed = True
        while True:
            message = await self._receive()
            if message["type"] == "http.request":
                body = message.get("body", b"")
                if body:
                    yield body
                if not message.get("more_body", False):
                    break
            elif message["type"] == "http.disconnect":
                self._is_disconnected = True
                raise ClientDisconnect()
        yield b""

    async def body(self) -> bytes:
        if not hasattr(self, "_body"):
            chunks = []
            async for chunk in self.stream():
                chunks.append(chunk)
            self._body = b"".join(chunks)
        return self._body

    async def json(self) -> Any:
        if not hasattr(self, "_json"):
            body = await self.body()
            self._json = json.loads(body)
        return self._json

    async def form(self) -> FormData:
        if not hasattr(self, "_form"):
            assert (
                parse_options_header is not None
            ), "The `python-multipart` library must be installed to use form parsing."
            content_type_header = self.headers.get("Content-Type")
            content_type, options = parse_options_header(content_type_header)
            if content_type == b"multipart/form-data":
                multipart_parser = MultiPartParser(self.headers, self.stream())
                self._form = await multipart_parser.parse()
            elif content_type == b"application/x-www-form-urlencoded":
                form_parser = FormParser(self.headers, self.stream())
                self._form = await form_parser.parse()
            else:
                self._form = FormData()
        return self._form

    async def close(self) -> None:
        if hasattr(self, "_form"):
            await self._form.close()

    async def is_disconnected(self) -> bool:
        if not self._is_disconnected:
            try:
                message = await asyncio.wait_for(self._receive(), timeout=0.0000001)
            except asyncio.TimeoutError:
                message = {}

            if message.get("type") == "http.disconnect":
                self._is_disconnected = True

        return self._is_disconnected

    async def send_push_promise(self, path: str) -> None:
        if "http.response.push" in self.scope.get("extensions", {}):
            raw_headers = []
            for name in SERVER_PUSH_HEADERS_TO_COPY:
                for value in self.headers.getlist(name):
                    raw_headers.append(
                        (name.encode("latin-1"), value.encode("latin-1"))
                    )
            await self._send(
                {"type": "http.response.push", "path": path, "headers": raw_headers}
            )



import os
import asyncio
import threading
import importlib
from types import ModuleType
from typing import Tuple, Dict, Any, Optional


class Singleton(type):
    def __init__(
        cls, name: str, bases: Tuple[type], namespace: Dict[str, Any],
    ) -> None:
        cls.instance = None
        super().__init__(name, bases, namespace)

    def __call__(cls, *args, **kwargs) -> Any:
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


def import_module(name: str) -> Optional[ModuleType]:
    """
    try importlib.import_module, nothing to do when module not be found.
    """
    if os.path.exists(os.path.join(os.getcwd(), name + ".py")) or os.path.exists(
        os.path.join(os.getcwd(), name, "__init__.py")
    ):
        return importlib.import_module(name)
    return None  # nothing to do when module not be found


class State(dict):
    """
    An object that can be used to store arbitrary state.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sync_lock = threading.Lock()
        self.async_lock = asyncio.Lock()

    def __enter__(self):
        self.sync_lock.acquire()
        return self

    def __exit__(self, exc_type, value, traceback):
        self.sync_lock.release()

    async def __aenter__(self):
        await self.async_lock.acquire()
        return self

    async def __aexit__(self, exc_type, value, traceback):
        self.async_lock.release()

    def __setattr__(self, name: Any, value: Any) -> None:
        self[name] = value

    def __getattr__(self, name: Any) -> Any:
        try:
            return self[name]
        except KeyError:
            message = "'{}' object has no attribute '{}'"
            raise AttributeError(message.format(self.__class__.__name__, name))

    def __delattr__(self, name: Any) -> None:
        del self[name]



import typing

from ..concurrency import keepasync
from .responses import Response, automatic
from .request import Request


MiddlewareMeta = keepasync("process_request", "process_response")


class MiddlewareMixin(metaclass=MiddlewareMeta):  # type: ignore

    mounts: typing.Sequence[typing.Callable] = ()

    def __init__(self, get_response: typing.Callable) -> None:
        self.get_response = self.mount_middleware(get_response)

    def mount_middleware(self, get_response: typing.Callable) -> typing.Callable:
        for middleware in reversed(self.mounts):
            get_response = middleware(get_response)
        return get_response

    async def __call__(self, request: Request) -> Response:
        response = await self.process_request(request)

        if response is None:
            response = await self.get_response(request)

        if isinstance(response, tuple):
            response = automatic(*response)
        else:
            response = automatic(response)

        response = await self.process_response(request, response)

        return response

    async def process_request(self, request: Request) -> typing.Union[None, typing.Any]:
        """Must return None, otherwise return the value as the result of this request."""

    async def process_response(self, request: Request, response: Response) -> Response:
        return response



"""
Load some preset functions into Index
"""
import os
import logging
from typing import Optional

from .config import here
from .autoreload import MonitorFile

logging.getLogger("multipart").setLevel(logging.INFO)


monitor: Optional[MonitorFile] = None


def check_on_startup() -> None:
    # monitor file event
    global monitor
    monitor = MonitorFile(here)


def clear_check_on_shutdown() -> None:
    global monitor
    if monitor is not None:
        monitor.stop()


def create_directories() -> None:
    """
    create directories for static & template
    """
    os.makedirs(os.path.join(here, "views"), exist_ok=True)
    os.makedirs(os.path.join(here, "static"), exist_ok=True)


def clear_directories() -> None:
    """
    if no files exist in the directory, delete them
    """

    def rmdir(dirpath: str) -> None:
        try:
            os.rmdir(dirpath)
        except OSError:
            pass

    rmdir(os.path.join(here, "views"))
    rmdir(os.path.join(here, "static"))



from copy import deepcopy
from typing import Any, List, Dict, Optional, Iterable, Union, Sequence

import yaml
from pydantic import BaseModel

from .types import File


def replace_definitions(schema: Dict[str, Any]) -> Dict[str, Any]:
    """
    replace $ref
    """
    schema = deepcopy(schema)

    if schema.get("definitions") is not None:

        def replace(value: Union[str, Sequence[Any], Dict[str, Any]]) -> None:
            if isinstance(value, str):
                return
            elif isinstance(value, Sequence):
                for _value in value:
                    replace(_value)
            elif isinstance(value, Dict):
                for _name in tuple(value.keys()):
                    if _name == "$ref":
                        define_schema = schema
                        for key in value["$ref"][2:].split("/"):
                            define_schema = define_schema[key]
                        # replace ref and del it
                        value.update(define_schema)
                        del value["$ref"]
                    elif isinstance(value[_name], Dict):
                        replace(value[_name])
                    elif isinstance(value[_name], Sequence):
                        for _value in value[_name]:
                            replace(_value)

        replace(schema["definitions"])
        replace(schema["properties"])
        del schema["definitions"]

    return schema


def schema_parameter(m: Optional[BaseModel], position: str) -> List[Dict[str, Any]]:
    """
    position: "path", "query", "header", "cookie"
    """
    result = []

    if m is not None:
        _schemas = replace_definitions(m.schema())
        properties: Dict[str, Any] = _schemas["properties"]
        _schemas["required"] = _schemas.get("required") or []

        for name, schema in properties.items():  # type: str, Dict[str, str]
            result.append(
                {
                    "in": position,
                    "name": name,
                    "description": schema.pop("description", ""),
                    "required": name in _schemas["required"],  # type: ignore
                    "schema": schema,
                }
            )
    return result


def schema_parameters(
    path: BaseModel = None,
    query: BaseModel = None,
    header: BaseModel = None,
    cookie: BaseModel = None,
) -> List[Dict[str, Any]]:

    return (
        schema_parameter(path, "path")
        + schema_parameter(query, "query")
        + schema_parameter(header, "header")
        + schema_parameter(cookie, "cookie")
    )


def schema_request_body(body: BaseModel = None) -> Optional[Dict[str, Any]]:
    if body is None:
        return None

    _schema = {"schema": replace_definitions(body.schema())}

    for field in body.__fields__.values():
        if issubclass(field.type_, File):
            return {
                "required": True,
                "content": {"multipart/form-data": _schema},
            }

    return {
        "required": True,
        "content": {"application/json": _schema},
    }


def schema_response(model: Union[BaseModel, str]) -> Dict[str, Any]:
    if isinstance(model, str):
        return yaml.safe_load(model.strip())
    return {"application/json": {"schema": replace_definitions(model.schema())}}



from types import TracebackType
from typing import Any, Type, Tuple, Union, Callable, Iterable, MutableMapping, Optional

from starlette.types import *

from .http.request import Request
from .http.responses import Response
from .websocket.request import WebSocket

__all__ = (
    "Request",
    "Response",
    "WebSocket",
    "Scope",
    "Message",
    "Receive",
    "Send",
    "ASGIApp",
    "ExcInfo",
    "Environ",
    "StartResponse",
    "WSGIApp",
)

# WSGI: view PEP3333
ExcInfo = Tuple[Type[BaseException], BaseException, Optional[TracebackType]]

Environ = MutableMapping[str, Any]

StartResponse = Callable[[str, Iterable[Tuple[str, str]], Optional[ExcInfo]], None]

WSGIApp = Callable[[Environ, StartResponse], Iterable[Union[str, bytes]]]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['indexpy', 'indexpy.http', 'indexpy.openapi', 'indexpy.websocket']

package_data = \
{'': ['*']}

install_requires = \
['a2wsgi>=0.3.4,<0.4.0',
 'aiofiles>=0.5.0,<0.6.0',
 'jinja2>=2.10.3,<3.0.0',
 'pydantic>=1.5,<2.0',
 'python-multipart>=0.0.5,<0.0.6',
 'pyyaml>=5.3,<6.0',
 'starlette>=0.13.1,<0.14.0',
 'uvicorn>=0.11.3,<0.12.0',
 'watchdog>=0.10.2,<0.11.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.7.4,<4.0.0'],
 ':python_version >= "3.6" and python_version < "3.7"': ['contextvars>=2.4,<3.0'],
 'gunicorn': ['gunicorn>=20.0.4,<21.0.0'],
 'pytest': ['pytest>=5.4.2,<6.0.0'],
 'requests': ['requests>=2.23.0,<3.0.0'],
 'test': ['requests>=2.23.0,<3.0.0', 'pytest>=5.4.2,<6.0.0']}

entry_points = \
{'console_scripts': ['index-cli = indexpy.cli:main']}

setup_kwargs = {
    'name': 'index.py',
    'version': '0.10.7',
    'description': 'An easy-to-use asynchronous web framework based on ASGI.',
    'long_description': '# index.py\n\n[![Github Action Test](https://github.com/abersheeran/index.py/workflows/Test/badge.svg)](https://github.com/abersheeran/index.py/actions?query=workflow%3ATest)\n[![PyPI](https://img.shields.io/pypi/v/index.py)](https://pypi.org/project/index.py/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/index.py)\n\n一个基于 ASGI 协议的高性能 web 框架。[Index.py 文档](https://abersheeran.github.io/index.py/)\n\n你也可以直接查看 [Example](https://github.com/abersheeran/index.py/tree/master/example) 来学习如何使用（文档偶尔会滞后一到两天，但 example 被纳入了自动化测试，所以会始终保持最新版）\n\n- 无需手动绑定路由 (文件系统映射URI)\n- 自动解析请求 & 生成文档 (基于 pydantic)\n- 可视化 API 接口 (基于 ReDoc, 针对中文字体优化)\n- 现代化的测试组件 (基于 pytest 与 requests)\n- 非常简单的部署 (基于 uvicorn 与 gunicorn)\n- 支持真正的热重载\n- 挂载 ASGI/WSGI 应用 (比 starlette 更易用)\n- 更好用的 background tasks\n- 可使用任何可用的 ASGI 生态\n\n## Install\n\n```bash\npip install -U index.py\n```\n\n或者直接从 Github 上安装最新版本（不稳定）\n\n```bash\npip install -U git+https://github.com/abersheeran/index.py\n```\n',
    'author': 'abersheeran',
    'author_email': 'me@abersheeran.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abersheeran/index.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import json
import typing

from starlette import status
from starlette.types import Message

from .request import WebSocket


class SocketView:

    encoding: typing.Optional[str] = None  # May be "text", "bytes", or "json".

    def __init__(self, websocket: WebSocket) -> None:
        self.websocket = websocket

    def __await__(self):
        return self.__call__().__await__()

    async def __call__(self) -> None:
        try:
            close_code = status.WS_1000_NORMAL_CLOSURE
            await self.on_connect()
            while True:
                message = await self.websocket.receive()
                if message["type"] == "websocket.receive":
                    data = await self.decode(message)
                    await self.on_receive(data)
                elif message["type"] == "websocket.disconnect":
                    close_code = int(message.get("code", status.WS_1000_NORMAL_CLOSURE))
                    break
        except Exception as exc:
            close_code = status.WS_1011_INTERNAL_ERROR
            raise exc from None
        finally:
            await self.on_disconnect(close_code)

    async def decode(self, message: Message) -> typing.Any:

        if self.encoding == "text":
            if "text" not in message:
                await self.websocket.close(code=status.WS_1003_UNSUPPORTED_DATA)
                raise RuntimeError("Expected text websocket messages, but got bytes")
            return message["text"]

        if self.encoding == "bytes":
            if "bytes" not in message:
                await self.websocket.close(code=status.WS_1003_UNSUPPORTED_DATA)
                raise RuntimeError("Expected bytes websocket messages, but got text")
            return message["bytes"]

        if self.encoding == "json":
            if message.get("text") is not None:
                text = message["text"]
            else:
                text = message["bytes"].decode("utf-8")

            try:
                return json.loads(text)
            except json.decoder.JSONDecodeError:
                await self.websocket.close(code=status.WS_1003_UNSUPPORTED_DATA)
                raise RuntimeError("Malformed JSON data received.")

        assert (
            self.encoding is None
        ), f"Unsupported 'encoding' attribute {self.encoding}"
        return message["text"] if message.get("text") else message["bytes"]

    async def on_connect(self) -> None:
        """Override to handle an incoming websocket connection"""
        await self.websocket.accept()

    async def on_receive(self, data: typing.Any) -> None:
        """Override to handle an incoming websocket message"""

    async def on_disconnect(self, close_code: int) -> None:
        """Override to handle a disconnecting websocket"""
        await self.websocket.close(code=close_code)



import typing
import functools
from inspect import signature

from pydantic import BaseModel, ValidationError

from ..concurrency import keepasync
from .responses import Response, automatic
from .request import Request


HTTP_METHOD_NAMES = [
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "head",
    "options",
    "trace",
]


class ViewMeta(keepasync(*HTTP_METHOD_NAMES)):  # type: ignore
    def __init__(
        cls,
        name: str,
        bases: typing.Tuple[type],
        namespace: typing.Dict[str, typing.Any],
    ):
        param_names = ["query", "header", "cookie", "body"]

        for key in filter(lambda key: key in HTTP_METHOD_NAMES, namespace.keys()):
            function = namespace[key]
            sig = signature(function)

            setattr(
                function,
                "__params__",
                {
                    param_name: sig.parameters[param_name].annotation
                    for param_name in param_names
                    if (
                        param_name in sig.parameters
                        and issubclass(sig.parameters[param_name].annotation, BaseModel)
                    )
                },
            )
        super().__init__(name, bases, namespace)


def merge_list(
    raw: typing.List[typing.Tuple[str, str]]
) -> typing.Dict[str, typing.Union[typing.List[str], str]]:
    """
    If there are values with the same key value, they are merged into a List.
    """
    d: typing.Dict[str, typing.Union[typing.List[str], str]] = {}
    for k, v in raw:
        if k in d:
            if isinstance(d[k], list):
                typing.cast(typing.List, d[k]).append(v)
            else:
                d[k] = [typing.cast(str, d[k]), v]
        else:
            d[k] = v
    return d


class HTTPView(metaclass=ViewMeta):  # type: ignore
    def __init__(self, request: Request) -> None:
        self.request = request

    def __await__(self):
        return self.__call__().__await__()

    @staticmethod
    async def partial(handler: typing.Callable, request: Request) -> typing.Any:
        __params__ = getattr(handler, "__params__", {})

        if not __params__:
            return handler

        params: typing.Dict[str, BaseModel] = {}

        # try to get parameters model and parse
        if "query" in __params__:
            params["query"] = __params__["query"](
                **merge_list(request.query_params.multi_items())
            )

        if "header" in __params__:
            params["header"] = __params__["header"](
                **merge_list(request.headers.items())
            )

        if "cookie" in __params__:
            params["cookie"] = __params__["cookie"](**request.cookies)

        # try to get body model and parse
        if "body" in __params__:
            if request.headers.get("Content-Type") == "application/json":
                _body_data = await request.json()
            else:
                _body_data = await request.form()
            params["body"] = __params__["body"](**_body_data)

        return functools.partial(handler, **params)

    async def __call__(self) -> Response:
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler. Also defer to the error handler if the
        # request method isn't on the approved list.

        if self.request.method.lower() in HTTP_METHOD_NAMES:
            handler = getattr(
                self, self.request.method.lower(), self.http_method_not_allowed
            )
        else:
            handler = self.http_method_not_allowed

        try:
            handler = await self.partial(handler, self.request)
        except ValidationError as e:
            resp = await self.catch_validation_error(e)
        else:
            resp = await handler()

        if isinstance(resp, tuple):
            resp = automatic(*resp)
        else:
            resp = automatic(resp)
        return resp

    async def catch_validation_error(
        self, e: ValidationError
    ) -> typing.Union[Response, tuple]:
        """
        Used to handle request parsing errors
        """
        return e.errors(), 400

    async def http_method_not_allowed(self) -> Response:
        return Response(
            status_code=405,
            headers={"Allow": ", ".join(self.allowed_methods()), "Content-Length": "0"},
        )

    async def options(self) -> Response:
        """Handle responding to requests for the OPTIONS HTTP verb."""
        return Response(
            headers={"Allow": ", ".join(self.allowed_methods()), "Content-Length": "0"}
        )

    @classmethod
    def allowed_methods(cls) -> typing.List[str]:
        return [m.upper() for m in HTTP_METHOD_NAMES if hasattr(cls, m)]
